package SHA2.SHA2;

import references.references.CharacterReference;

import static Bits.Bitwise.Bitwise.*;
import static IntegerArithmetic.U32Arithmetic.U32Arithmetic.ConvertToU32;
import static aarrays.arrays.arrays.aStringToNumberArray;
import static charCharacters.Characters.Characters.charToLowerCase;
import static java.lang.Math.*;

import static aarrays.arrays.arrays.aNumberArrayToString;
import static numbers.NumberToString.NumberToString.GetSingleDigitCharacterFromNumberWithCheck;

public class SHA2 {

    public static double WordFromBytes(double d1, double d2, double d3, double d4) {
        return d1 * (256d * 256d * 256d) + d2 * (256d * 256d) + d3 * 256d + d4;
    }

    public static double ExtractFirstByte(double n) {
        return n % 256d;
    }

    public static double AddUnsignedInteger4ByteModular(double a, double b) {
        return ConvertToU32(a+b);
    }

    public static double SHA256SIG0(double w) {
        double n;

        n = RotateRight4Bytes(w, 7d);
        n = Xor4Byte(n, RotateRight4Bytes(w, 18d));
        n = Xor4Byte(n, ShiftRight4Byte(w, 3d));

        return n;
    }

    public static double SHA256SIG1(double w) {
        double n;

        n = RotateRight4Bytes(w, 17d);
        n = Xor4Byte(n, RotateRight4Bytes(w, 19d));
        n = Xor4Byte(n, ShiftRight4Byte(w, 10d));

        return n;
    }

    public static double SHA256CH(double x, double y, double z) {
        double n;

        n = And4Byte(x, y);
        n = Xor4Byte(n, And4Byte(Not4Byte(x), z));

        return n;
    }

    public static double SHA256MAJ(double x, double y, double z) {
        double n;

        n = Xor4Byte(And4Byte(x, y), And4Byte(x, z));
        n = Xor4Byte(n, And4Byte(y, z));

        return n;
    }

    public static double SHA256EP0(double w) {
        double n;

        n = RotateRight4Bytes(w, 2d);
        n = Xor4Byte(n, RotateRight4Bytes(w, 13d));
        n = Xor4Byte(n, RotateRight4Bytes(w, 22d));

        return n;
    }

    public static double SHA256EP1(double w) {
        double n;

        n = RotateRight4Bytes(w, 6d);
        n = Xor4Byte(n, RotateRight4Bytes(w, 11d));
        n = Xor4Byte(n, RotateRight4Bytes(w, 25d));

        return n;
    }

    public static void SHA256Transform(double[] chunk, double[] state, double[] k) {
        double[] m;
        double a, b, c, d, e, f, g, h, t1, t2;

        double i, j, r;

        m = new double[(int) 64d];

        j = 0;
        for (i = 0d; i < 16d; i = i + 1d) {
            r = chunk[(int) j] * 256d * 256d * 256d;
            r = AddUnsignedInteger4ByteModular(r, chunk[(int) (j + 1d)] * 256d * 256d);
            r = AddUnsignedInteger4ByteModular(r, chunk[(int) (j + 2d)] * 256d);
            r = AddUnsignedInteger4ByteModular(r, chunk[(int) (j + 3d)]);
            m[(int) i] = r;

            j = j + 4d;
        }

        for (; i < 64d; i = i + 1d) {
            r = SHA256SIG1(m[(int) (i - 2d)]);
            r = AddUnsignedInteger4ByteModular(r, m[(int) (i - 7d)]);
            r = AddUnsignedInteger4ByteModular(r, SHA256SIG0(m[(int) (i - 15d)]));
            r = AddUnsignedInteger4ByteModular(r, m[(int) (i - 16d)]);
            m[(int) i] = r;
        }

        a = state[(int) 0d];
        b = state[(int) 1d];
        c = state[(int) 2d];
        d = state[(int) 3d];
        e = state[(int) 4d];
        f = state[(int) 5d];
        g = state[(int) 6d];
        h = state[(int) 7d];

        for (i = 0d; i < 64d; i = i + 1d) {
            t1 = h;
            t1 = AddUnsignedInteger4ByteModular(t1, SHA256EP1(e));
            t1 = AddUnsignedInteger4ByteModular(t1, SHA256CH(e, f, g));
            t1 = AddUnsignedInteger4ByteModular(t1, k[(int) i]);
            t1 = AddUnsignedInteger4ByteModular(t1, m[(int) i]);

            t2 = SHA256EP0(a);
            t2 = AddUnsignedInteger4ByteModular(t2, SHA256MAJ(a, b, c));

            h = g;
            g = f;
            f = e;
            e = AddUnsignedInteger4ByteModular(d, t1);
            d = c;
            c = b;
            b = a;
            a = AddUnsignedInteger4ByteModular(t1, t2);
        }

        state[(int) 0d] = AddUnsignedInteger4ByteModular(state[(int) 0d], a);
        state[(int) 1d] = AddUnsignedInteger4ByteModular(state[(int) 1d], b);
        state[(int) 2d] = AddUnsignedInteger4ByteModular(state[(int) 2d], c);
        state[(int) 3d] = AddUnsignedInteger4ByteModular(state[(int) 3d], d);
        state[(int) 4d] = AddUnsignedInteger4ByteModular(state[(int) 4d], e);
        state[(int) 5d] = AddUnsignedInteger4ByteModular(state[(int) 5d], f);
        state[(int) 6d] = AddUnsignedInteger4ByteModular(state[(int) 6d], g);
        state[(int) 7d] = AddUnsignedInteger4ByteModular(state[(int) 7d], h);
    }

    public static char HexDigitToLowerCaseCharacter(double b) {
        CharacterReference c;
        char ch;

        c = new CharacterReference();
        GetSingleDigitCharacterFromNumberWithCheck(b, 16d, c);
        ch = c.characterValue;
        delete(c);

        return charToLowerCase(ch);
    }

    public static char[] SHA256(double[] data) {
        double i, j; // Used for indexing

        double[] state; // Used to store the initial hashes and for transforming
        double[] k; // Stores 64 4-byte sized constants that are used in the algorithm

        double datalen;
        double bitlen;

        double[] chunk; // 64 bytes used for each chunk

        double[] hash; // The result
        char[] hashBytes;
        double b;

        char[] result; // Converted into hex string form (stores 64 characters)

        datalen = 0d;
        bitlen = 0d;

        // Init constants
        state = new double[(int) (8d)];
        state[(int) 0d] = WordFromBytes(106d, 9d, 230d, 103d);
        state[(int) 1d] = WordFromBytes(187d, 103d, 174d, 133d);
        state[(int) 2d] = WordFromBytes(60d, 110d, 243d, 114d);
        state[(int) 3d] = WordFromBytes(165d, 79d, 245d, 58d);
        state[(int) 4d] = WordFromBytes(81d, 14d, 82d, 127d);
        state[(int) 5d] = WordFromBytes(155d, 5d, 104d, 140d);
        state[(int) 6d] = WordFromBytes(31d, 131d, 217d, 171d);
        state[(int) 7d] = WordFromBytes(91d, 224d, 205d, 25d);

        k = new double[(int) (64d)];
        k[(int) 0d] = WordFromBytes(66d, 138d, 47d, 152d);
        k[(int) 1d] = WordFromBytes(113d, 55d, 68d, 145d);
        k[(int) 2d] = WordFromBytes(181d, 192d, 251d, 207d);
        k[(int) 3d] = WordFromBytes(233d, 181d, 219d, 165d);
        k[(int) 4d] = WordFromBytes(57d, 86d, 194d, 91d);
        k[(int) 5d] = WordFromBytes(89d, 241d, 17d, 241d);
        k[(int) 6d] = WordFromBytes(146d, 63d, 130d, 164d);
        k[(int) 7d] = WordFromBytes(171d, 28d, 94d, 213d);
        k[(int) 8d] = WordFromBytes(216d, 7d, 170d, 152d);
        k[(int) 9d] = WordFromBytes(18d, 131d, 91d, 1d);
        k[(int) 10d] = WordFromBytes(36d, 49d, 133d, 190d);
        k[(int) 11d] = WordFromBytes(85d, 12d, 125d, 195d);
        k[(int) 12d] = WordFromBytes(114d, 190d, 93d, 116d);
        k[(int) 13d] = WordFromBytes(128d, 222d, 177d, 254d);
        k[(int) 14d] = WordFromBytes(155d, 220d, 6d, 167d);
        k[(int) 15d] = WordFromBytes(193d, 155d, 241d, 116d);
        k[(int) 16d] = WordFromBytes(228d, 155d, 105d, 193d);
        k[(int) 17d] = WordFromBytes(239d, 190d, 71d, 134d);
        k[(int) 18d] = WordFromBytes(15d, 193d, 157d, 198d);
        k[(int) 19d] = WordFromBytes(36d, 12d, 161d, 204d);
        k[(int) 20d] = WordFromBytes(45d, 233d, 44d, 111d);
        k[(int) 21d] = WordFromBytes(74d, 116d, 132d, 170d);
        k[(int) 22d] = WordFromBytes(92d, 176d, 169d, 220d);
        k[(int) 23d] = WordFromBytes(118d, 249d, 136d, 218d);
        k[(int) 24d] = WordFromBytes(152d, 62d, 81d, 82d);
        k[(int) 25d] = WordFromBytes(168d, 49d, 198d, 109d);
        k[(int) 26d] = WordFromBytes(176d, 3d, 39d, 200d);
        k[(int) 27d] = WordFromBytes(191d, 89d, 127d, 199d);
        k[(int) 28d] = WordFromBytes(198d, 224d, 11d, 243d);
        k[(int) 29d] = WordFromBytes(213d, 167d, 145d, 71d);
        k[(int) 30d] = WordFromBytes(6d, 202d, 99d, 81d);
        k[(int) 31d] = WordFromBytes(20d, 41d, 41d, 103d);
        k[(int) 32d] = WordFromBytes(39d, 183d, 10d, 133d);
        k[(int) 33d] = WordFromBytes(46d, 27d, 33d, 56d);
        k[(int) 34d] = WordFromBytes(77d, 44d, 109d, 252d);
        k[(int) 35d] = WordFromBytes(83d, 56d, 13d, 19d);
        k[(int) 36d] = WordFromBytes(101d, 10d, 115d, 84d);
        k[(int) 37d] = WordFromBytes(118d, 106d, 10d, 187d);
        k[(int) 38d] = WordFromBytes(129d, 194d, 201d, 46d);
        k[(int) 39d] = WordFromBytes(146d, 114d, 44d, 133d);
        k[(int) 40d] = WordFromBytes(162d, 191d, 232d, 161d);
        k[(int) 41d] = WordFromBytes(168d, 26d, 102d, 75d);
        k[(int) 42d] = WordFromBytes(194d, 75d, 139d, 112d);
        k[(int) 43d] = WordFromBytes(199d, 108d, 81d, 163d);
        k[(int) 44d] = WordFromBytes(209d, 146d, 232d, 25d);
        k[(int) 45d] = WordFromBytes(214d, 153d, 6d, 36d);
        k[(int) 46d] = WordFromBytes(244d, 14d, 53d, 133d);
        k[(int) 47d] = WordFromBytes(16d, 106d, 160d, 112d);
        k[(int) 48d] = WordFromBytes(25d, 164d, 193d, 22d);
        k[(int) 49d] = WordFromBytes(30d, 55d, 108d, 8d);
        k[(int) 50d] = WordFromBytes(39d, 72d, 119d, 76d);
        k[(int) 51d] = WordFromBytes(52d, 176d, 188d, 181d);
        k[(int) 52d] = WordFromBytes(57d, 28d, 12d, 179d);
        k[(int) 53d] = WordFromBytes(78d, 216d, 170d, 74d);
        k[(int) 54d] = WordFromBytes(91d, 156d, 202d, 79d);
        k[(int) 55d] = WordFromBytes(104d, 46d, 111d, 243d);
        k[(int) 56d] = WordFromBytes(116d, 143d, 130d, 238d);
        k[(int) 57d] = WordFromBytes(120d, 165d, 99d, 111d);
        k[(int) 58d] = WordFromBytes(132d, 200d, 120d, 20d);
        k[(int) 59d] = WordFromBytes(140d, 199d, 2d, 8d);
        k[(int) 60d] = WordFromBytes(144d, 190d, 255d, 250d);
        k[(int) 61d] = WordFromBytes(164d, 80d, 108d, 235d);
        k[(int) 62d] = WordFromBytes(190d, 249d, 163d, 247d);
        k[(int) 63d] = WordFromBytes(198d, 113d, 120d, 242d);

        // Begin algorithm
        chunk = new double[(int) 64d];

        for (i = 0d; i < data.length; i = i + 1d) {
            chunk[(int) datalen] = data[(int) i];
            datalen = datalen + 1d;

            if (datalen == 64d) {
                SHA256Transform(chunk, state, k);
                bitlen = bitlen + 512d;
                datalen = 0d;
            }
        }

        // Final steps
        i = datalen;

        if (datalen < 56d) {
            chunk[(int) i] = 128d;
            i = i + 1d;

            for (; i < 56d; i = i + 1d) {
                chunk[(int) i] = 0d;
            }
        } else {
            chunk[(int) i] = 128d;
            i = i + 1d;

            for (; i < 64d; i = i + 1d) {
                chunk[(int) i] = 0d;
            }

            SHA256Transform(chunk, state, k);

            for (j = 0d; j < 56d; j = j + 1d) {
                chunk[(int) j] = 0d;
            }
        }

        // Append to the padding the total message's length in bits and transform.
        bitlen = bitlen + datalen * 8d;
        chunk[63] = ExtractFirstByte(bitlen);
        chunk[62] = ExtractFirstByte(ShiftRight4Byte(bitlen, 8d));
        chunk[61] = ExtractFirstByte(ShiftRight4Byte(bitlen, 16d));
        chunk[60] = ExtractFirstByte(ShiftRight4Byte(bitlen, 24d));
        chunk[59] = ExtractFirstByte(ShiftRight4Byte(bitlen, 32d));
        chunk[58] = ExtractFirstByte(ShiftRight4Byte(bitlen, 40d));
        chunk[57] = ExtractFirstByte(ShiftRight4Byte(bitlen, 48d));
        chunk[56] = ExtractFirstByte(ShiftRight4Byte(bitlen, 56d));
        SHA256Transform(chunk, state, k);

        hash = new double[(int) 32d];

        // Since this implementation uses little endian byte ordering and SHA uses big
        // endian,
        // reverse all the bytes when copying the final state to the output hash.
        for (i = 0d; i < 4d; i = i + 1d) {
            hash[(int) (i)] = ExtractFirstByte(ShiftRight4Byte(state[(int) 0d], 24d - i * 8d));
            hash[(int) (i + 4d)] = ExtractFirstByte(ShiftRight4Byte(state[(int) 1d], 24d - i * 8d));
            hash[(int) (i + 8d)] = ExtractFirstByte(ShiftRight4Byte(state[(int) 2d], 24d - i * 8d));
            hash[(int) (i + 12d)] = ExtractFirstByte(ShiftRight4Byte(state[(int) 3d], 24d - i * 8d));
            hash[(int) (i + 16d)] = ExtractFirstByte(ShiftRight4Byte(state[(int) 4d], 24d - i * 8d));
            hash[(int) (i + 20d)] = ExtractFirstByte(ShiftRight4Byte(state[(int) 5d], 24d - i * 8d));
            hash[(int) (i + 24d)] = ExtractFirstByte(ShiftRight4Byte(state[(int) 6d], 24d - i * 8d));
            hash[(int) (i + 28d)] = ExtractFirstByte(ShiftRight4Byte(state[(int) 7d], 24d - i * 8d));
        }

        hashBytes = aNumberArrayToString(hash);

        // Convert into hex string form
        result = new char[(int) 64d];
        for (i = 0d; i < 32d; i = i + 1d) {
            b = hashBytes[(int) i];

            result[(int) (i * 2d)] = HexDigitToLowerCaseCharacter(floor(b / 16d));
            result[(int) (i * 2d + 1d)] = HexDigitToLowerCaseCharacter(b - floor(b / 16d) * 16d);
        }

        return result;
    }

    public static char[] SHA256FromString(char [] str) {
        double[] d;
        char[] s;

        d = aStringToNumberArray(str);
        s = SHA256(d);

        return s;
    }
}